/*
 * uibottom.h - 3DS bottom screen handling incl. virtual keyboard
 *
 * Written by
 *  Sebastian Weber <me@badda.de>
 *
 * This file is part of VICE3DS
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#ifndef VICE_UIBOTTOM_H
#define VICE_UIBOTTOM_H

#include "uimenu.h"

// exposed definitions
typedef struct {
	int x,y,w,h,key,shift,sticky,flags;
	char *name;
} uikbd_key;

enum bottom_action {
	UIB_NO = 0,
	UIB_REPAINT			=		0x01,	// 0x01
	UIB_RECALC_KEYPRESS =		0x03,	// 0x02
	UIB_RECALC_KEYBOARD =		0x07,	// 0x02
	UIB_RECALC_SBUTTONS =		0x0b,	// 0x04
	UIB_ALL =					0x0f,
	UIB_GET_REPAINT =			0x01,
	UIB_GET_RECALC_KEYPRESS =	0x02,
	UIB_GET_RECALC_KEYBOARD =	0x04,
	UIB_GET_RECALC_SBUTTONS =	0x08,
	UIB_RECALC_MENU =			0x10,
	UIB_RECALC_GB64 =			0x20
};

enum str_alignment {
	ALIGN_LEFT,
	ALIGN_RIGHT,
	ALIGN_CENTER
};

enum font_size {
	FONT_SMALL,
	FONT_MEDIUM,
	FONT_BIG,
	FONT_SYM
};

// exposed functions
extern void sdl_uibottom_draw(void);
extern ui_menu_action_t sdl_uibottom_mouseevent(SDL_Event *);
extern void toggle_keyboard(void);
extern void setBottomBacklight (int on);
extern int getBottomBacklight();
extern int is_keyboard_hidden();
extern int uibottom_resources_init();
extern void uibottom_resources_shutdown();
extern int uibottom_editmode_is_on();
extern void uibottom_toggle_editmode();
extern void toggle_help(int inmenu);
extern void toggle_menu(int active, ui_menu_entry_t *item);
extern void uib_printtext(SDL_Surface *s, const char *str, int xo, int yo, int w, int h, enum font_size size, SDL_Color col);
extern void uib_printstring(SDL_Surface *s, const char *str, int x, int y, int maxchars, enum str_alignment align, enum font_size size, SDL_Color col);
extern void uib_show_message(u32 ms_time, char *format, ... );
extern void uibottom_unload_stuff();

// exposed variables
extern uikbd_key *uikbd_keypos;
extern volatile enum bottom_action uibottom_must_redraw;
extern int help_on;
extern int grab_sbuttons;
extern SDL_Event lastevent;
extern int menu_invisible;
extern SDL_Surface *mediumchars;

#endif
